/*
  aiplayer.cpp - computer KBalm player

  Copyright 2004 by Richard Smith <kde@metafoo.co.uk>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
*/

#include "aiplayer.h"

#include "board.h"
#include "game.h"

#include <klocale.h>

#include <qtimer.h>

namespace KBalmAI
{

double boardValue( const Balm::Board &board, int player )
{
	return (double)board.score( player ) / (double)board.totalScore();
}

//BEGIN class Turn

Turn::Turn( Player &player, Balm::Game &game )
	: Balm::Turn( player, game )
{
	debug() << endl << k_funcinfo << "new AI turn for player " << player.number() << endl;
	setStatus( i18n( "Thinking..." ) );
	
	int depth = player.skill();
	for ( int x = 0; x < board().width(); ++x )
		for ( int y = 0; y < board().height(); ++y )
			if ( board().at( x, y )->owner() == &player )
			{
				//debug() << "found possible at (" << x << ", " << y << ")" << endl;
				_moves.push_back( PossibleMove( x, y, new Thought( depth, board(), game, player.number(), player.number(), x, y ) ) );
			}
	
	QTimer::singleShot( 20000, this, SLOT( bored() ) );
	
	think();
}

Turn::~Turn()
{
	while ( !_moves.empty() )
	{
		delete _moves.front().thought;
		_moves.pop_front();
	}
	debug() << k_funcinfo << "AI thinking finished." << endl;
}

void Turn::think()
{
	bool done = true;
	
	std::pair<int,int> best;
	double bestScore = -1.0;
	
	for ( _moves_t::iterator it = _moves.begin(); it != _moves.end(); ++it )
	{
		if ( it->thought->think() )
		{
			double value = it->thought->value();
			if ( value > bestScore )
			{
				bestScore = value;
				best = it->move;
			}
		}
		else
		{
			done = false;
		}
	}
	
	if ( done )
		move( best.first, best.second );
	else
		QTimer::singleShot( 0, this, SLOT( think() ) );
}

void Turn::bored()
{
	debug() << k_funcinfo << " timed out calculating best move" << endl;
	setStatus( i18n( "Got bored" ) );
	
	std::pair<int,int> best;
	double bestScore = -1.0;
	
	for ( _moves_t::iterator it = _moves.begin(); it != _moves.end(); ++it )
	{
		it->thought->bored();
		double value = it->thought->value();
		if ( value > bestScore )
		{
			bestScore = value;
			best = it->move;
		}
	}
	
	move( best.first, best.second );
}

//END class Turn


//BEGIN class Thought

int thoughts;

Thought::Thought( int level, Balm::Board &board, Balm::Game &game, int turn, int self, int x, int y )
	: _level(level), _board(board), _game(game), _turn(turn), _self(self), _done(false)
{
	_board.rotate( x, y );
	_infect.reset( new Balm::InfectBase( _board, x, y ) );
	++thoughts;
	if ( (thoughts % 1000) == 0 )
		debug() << k_funcinfo << "Now " << thoughts << " thoughts." << endl;
}

Thought::~Thought()
{
	while ( !_thoughts.empty() )
	{
		delete _thoughts.front();
		_thoughts.pop_front();
	}
	--thoughts;
}

bool Thought::think()
{
	if ( _done )
		return true;

	if ( _infect.get() )
	{
		// still infecting?
		if ( !_infect->process() )
			return false;
		_infect.reset();
		
		if ( _level == 0 || thoughts > 100000 )
		{
			_value = boardValue( _board, _self );
			_done = true;
			return true;
		}
		
		int next = _turn;
		do
		{
			if ( ++next >= _game.players() )
				next = 0;
		} while ( _board.score( next ) == 0 );
		
		// only player left?
		if ( next == _turn )
		{
			_value = (_turn == _self) ? 1.0 : 0.0;
			_done = true;
			return true;
		}
		
		Balm::Player *player = _game.player( next );
		//debug() << k_funcinfo << "finished infecting for level " << _level << ". player = " << player << " for turn " << _turn << endl;
		for ( int x = 0; x < _board.width(); ++x )
			for ( int y = 0; y < _board.height(); ++y )
				if ( _board.at( x, y )->owner() == player )
					_thoughts.push_back( new Thought( _level - 1, _board, _game, next, _self, x, y ) );
	}
	
	bool done = true;
	
	for ( _thoughts_t::iterator it = _thoughts.begin(); it != _thoughts.end(); ++it )
	{
		if ( !(*it)->think() )
			//return false;
			done = false;
	}
	
	if ( done )
	{
		calculateValue();
		_done = true;
		
		while ( !_thoughts.empty() )
		{
			delete _thoughts.front();
			_thoughts.pop_front();
		}
	}
	
	return done;
}

void Thought::calculateValue()
{
	int next = _turn;
	do
	{
		if ( ++next >= _game.players() )
			next = 0;
	} while ( _board.score( next ) == 0 );
	
	// game over?
	if ( next == _turn )
	{
		if ( _turn == _self )
			_value = 1.0;
		else
			_value = 0.0;
		return;
	}
	
	// if we go next, we'll choose the best option
	if ( next == _self )
	{
		_value = 0.0;
		for ( _thoughts_t::iterator it = _thoughts.begin(); it != _thoughts.end(); ++it )
			if ( (*it)->value() > _value )
				_value = (*it)->value();
	}
	// if opponent goes next, assume they'll choose the worst for us
	else
	{
		_value = 1.0;
		for ( _thoughts_t::iterator it = _thoughts.begin(); it != _thoughts.end(); ++it )
			if ( (*it)->value() < _value )
				_value = (*it)->value();
	}
	//debug() << "finished thinking, value = " << _value << endl;
}

void Thought::bored()
{
	for ( _thoughts_t::iterator it = _thoughts.begin(); it != _thoughts.end(); ++it )
		(*it)->bored();
	
	if ( _infect.get() )
		_value = boardValue( _board, _self );
	else
		calculateValue();
	
	_done = true;
}

double Thought::value()
{
	return _value;
}

//END class Thought

} // namespace KBalmAIPlayer

#include "aiplayer.moc"
//kate: indent-mode: csands;
