/*
  board.cpp - KBalm game board
  
  Copyright 2004 by Richard Smith <kde@metafoo.co.uk>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
*/

#include "board.h"

#include "player.h"

#include <kdebug.h>

#include <cstdlib>

namespace Balm
{

//BEGIN class Direction

Direction::Direction( Direction::RandomTag )
	: _name( static_cast<Direction::Name>(std::rand() % 4) )
{
}

int Direction::x() const
{
	switch ( name() )
	{
		case Up: case Down: return 0;
		case Left: return -1;
		case Right: return 1;
	}
	error() << "invalid direction name " << name() << endl;
	return 0;
}

int Direction::y() const
{
	switch ( name() )
	{
		case Left: case Right: return 0;
		case Up: return -1;
		case Down: return 1;
	}
	error() << "invalid direction name " << name() << endl;
	return 0;
}
	
Direction Direction::rotate() const
{
	switch ( name() )
	{
		case Left: return Up;
		case Up: return Right;
		case Right: return Down;
		case Down: return Left;
	}
	error() << "invalid direction name " << name() << endl;
	return Random;
}

//END class Direction


//BEGIN class Node

//END class Node


//BEGIN class Board

Board::Board( const Board &o )
	: _nodes( o._nodes ), _score( o._score )
{
}

Board::Board( int width, int height, std::auto_ptr<BoardSignals> signalz )
	: _signals( signalz )
{
	_nodes.resize( width );
	for ( uint x = 0; x < width; ++x )
	{
		_nodes[x].reserve( height );
		for ( uint y = 0; y < height; ++y )
			_nodes[x].push_back( Node( Direction::Random ) );
	}
}

Board::~Board()
{
}

int Board::width() const
{
	return _nodes.size();
}

int Board::height() const
{
	return _nodes[0].size();
}

BoardSignals *Board::signalz() const
{
	return _signals.get();
}

int Board::score( int player ) const
{
	if ( player < 0 || player >= _score.size() )
		return 0;
	return _score[player];
}

int Board::totalScore() const
{
	int score = 0;
	for ( uint n = 0; n < _score.size(); ++n )
		score += _score[n];
	return score;
}

const Node *Board::at( int x, int y ) const
{
	if ( x < 0 || y < 0 || x >= width() || y >= height() )
	{
		//debug() << k_funcinfo << "position (" << x << ", " << y << ") out of range" << endl;
		return 0;
	}
	return &_nodes[x][y];
}

void Board::setOwner( int x, int y, const Player *owner )
{
	if ( x < 0 || y < 0 || x >= width() || y >= height() )
	{
		debug() << k_funcinfo << "position (" << x << ", " << y << ") out of range" << endl;
		return;
	}
	
	Node &node = _nodes[x][y];
	
	if ( node.owner() == owner )
		return;
	
	if ( owner && owner->number() >= _score.size() )
		_score.resize( owner->number() + 1 );
	if ( node.owner() )
		_score[ node.owner()->number() ]--;
	if ( owner )
		_score[ owner->number() ]++;
	
	node.setOwner( owner );
	if ( _signals.get() )
	{
		emit _signals->scoreChanged();
		emit _signals->nodeChanged( x, y );
	}
}

void Board::setDirection( int x, int y, Direction dir )
{
	if ( x < 0 || y < 0 || x >= width() || y >= height() )
	{
		debug() << k_funcinfo << "position (" << x << ", " << y << ") out of range" << endl;
		return;
	}
	
	Node &node = _nodes[x][y];
	
	if ( node.direction() == dir )
		return;
	
	node.setDirection( dir );
	if ( _signals.get() )
		emit _signals->nodeChanged( x, y );
}

void Board::rotate( int x, int y )
{
	if ( x < 0 || y < 0 || x >= width() || y >= height() )
	{
		debug() << k_funcinfo << "position (" << x << ", " << y << ") out of range" << endl;
		return;
	}
	
	Node &node = _nodes[x][y];
	node.setDirection( node.direction().rotate() );
	if ( _signals.get() )
		emit _signals->nodeChanged( x, y );
}

bool Board::connected( int x, int y, Direction dir ) const
{
	if ( at( x, y ) && at( x, y )->direction() == dir )
		return true;
	if ( at( x + dir.x(), y + dir.y() ) && at( x + dir.x(), y + dir.y() )->direction() == dir.rotate().rotate() )
		return true;
	return false;
}

QString Board::playerNumber( int x, int y ) const
{
	if ( at( x, y ) && at( x, y )->owner() )
		return QString::number( at( x, y )->owner()->number() + 1 );
	return "0";
}

BoardSignals::BoardSignals() : QObject(0) {}

} // namespace Balm

#include "board.moc"
//kate: indent-mode: csands;
