/*
  boardwidget.cpp - widget displaying a KBalm Board
  
  Copyright 2004 by Richard Smith <kde@metafoo.co.uk>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
*/

#include "boardwidget.h"

#include "board.h"
#include "pixmapcache.h"
#include "player.h"

#include <qpainter.h>
#include <qpixmap.h>

#include <utility>

namespace Balm
{

const int NODE_WIDTH = 22;
const int NODE_HEIGHT = 22;

BoardWidget::BoardWidget( QWidget *parent, Board *board, const char *name )
	: QScrollView( parent, name ), _board( board ), _showSel( false ), _selX( 0 ), _selY( 0 ), _cache( new KBalmPixmapCache )
{
	const int width = board->width() * NODE_WIDTH, height = board->height() * NODE_HEIGHT;
	resizeContents( width, height );
	setMaximumSize( width + 2 * frameWidth(), height + 2 * frameWidth() );
	
	// evil
	class FixQScrollView : public QWidget { public: void setFlags( uint flags ) { setWFlags( flags ); } };
	static_cast<FixQScrollView*>(viewport())->setFlags( WNoAutoErase | WStaticContents );
	setWFlags( WNoAutoErase | WStaticContents );
	
	connect( board->signalz(), SIGNAL( nodeChanged(int,int) ), SLOT( boardChanged(int,int) ) );
}

BoardWidget::~BoardWidget()
{
}


void BoardWidget::moveLeft()
{
	moveSelection( -1, 0 );
}

void BoardWidget::moveRight()
{
	moveSelection( 1, 0 );
}

void BoardWidget::moveUp()
{
	moveSelection( 0, -1 );
}

void BoardWidget::moveDown()
{
	moveSelection( 0,  1 );
}

void BoardWidget::moveSelection( int xOff, int yOff )
{
	bool changed = false;
	int oldX = _selX, oldY = _selY;
	if ( 0 <= _selX + xOff && _selX + xOff < _board->width() )
	{
		changed = true;
		_selX += xOff;
	}
	if ( 0 <= _selY + yOff && _selY + yOff < _board->height() )
	{
		changed = true;
		_selY += yOff;
	}
	
	// hide old selection marker
	if ( changed && _showSel )
	{
		// must paint now - if updateContents is used and then the view is scrolled
		// by showSelection, the redraw doesn't happen due to a Qt bug.
		repaintContents( oldX * NODE_WIDTH, oldY * NODE_HEIGHT, NODE_WIDTH, NODE_HEIGHT, false );
	}
	
	// show current selection marker
	showSelection();
	boardChanged( _selX, _selY );
}

void BoardWidget::showSelection()
{
	if ( !_showSel )
	{
		_showSel = true;
		boardChanged( _selX, _selY );
	}
	ensureNodeVisible( _selX, _selY );
}

void BoardWidget::hideSelection()
{
	if ( _showSel )
	{
		_showSel = false;
		boardChanged( _selX, _selY );
	}
}

void BoardWidget::selectCurrent()
{
	showSelection();
	emit nodeSelected( _selX, _selY );
}


void BoardWidget::drawHConnect( QPainter *p, int x, int y, const QString &player )
{
	if ( !_board->connected( x, y, Direction::Right ) )
		return;
		
	const int connectLength = 8, connectWidth = 4;
	QPixmap connect = _cache->pixmap( "pics/connect_" + player + ".png" );
	p->drawPixmap( (x + 1) * NODE_WIDTH - connectLength / 2, y * NODE_HEIGHT + (NODE_HEIGHT - connectWidth) / 2, connect );
}

void BoardWidget::drawVConnect( QPainter *p, int x, int y, const QString &player )
{
	if ( !_board->connected( x, y, Direction::Down ) )
		return;
	
	const int connectLength = 8, connectWidth = 4;
	QPixmap connect = _cache->pixmap( "pics/connect_" + player + ".png" );
	
	p->save();
	p->translate( x * NODE_WIDTH + (NODE_WIDTH + connectWidth) / 2, (y + 1) * NODE_HEIGHT - connectLength / 2 );
	p->rotate( 90.0 );
	p->drawPixmap( 0, 0, connect );
	p->restore();
}

void BoardWidget::drawContents( QPainter *q, int clipX, int clipY, int clipW, int clipH )
{
	// yet another Qt bug...
	if ( clipW <= 0 || clipH <= 0 )
		return;
	
	QPixmap arrow = _cache->pixmap( "pics/arrow.png" );
	
	QPixmap back( clipW, clipH );
	QPainter painter( &back );
	painter.translate( -clipX, -clipY );
	QPainter *p = &painter;
	
	QColor background("gray");
	p->fillRect( clipX, clipY, clipW, clipH, background );
	
	const uint startX = clipX / NODE_WIDTH, startY = clipY / NODE_HEIGHT,
	           w = std::min( _board->width(), (clipX + clipW + NODE_WIDTH - 1) / NODE_WIDTH ),
	           h = std::min( _board->height(), (clipY + clipH + NODE_HEIGHT - 1) / NODE_HEIGHT );
	
	// draw shadows
	p->save();
	p->translate( 2, 2 );
	for ( uint x = startX; x < w; ++x )
		drawVConnect( p, x, startY - 1, "shadow" );
	for ( uint y = startY; y < h; ++y )
		drawHConnect( p, startX - 1, y, "shadow" );
	for ( uint x = startX; x < w; ++x )
	{
		for ( uint y = startY; y < h; ++y )
		{
			QPixmap shadow = _cache->pixmap( "pics/node_shadow.png" );
			p->drawPixmap( x * NODE_WIDTH, y * NODE_HEIGHT, shadow );
			
			drawHConnect( p, x, y, "shadow" );
			drawVConnect( p, x, y, "shadow" );
		}
	}
	p->restore();
	
	// draw stuff
	for ( uint x = startX; x < w; ++x )
		drawVConnect( p, x, startY - 1, _board->playerNumber( x, startY ) );
	for ( uint y = startY; y < h; ++y )
		drawHConnect( p, startX - 1, y, _board->playerNumber( startX, y ) );
	for ( uint x = startX; x < w; ++x )
	{
		for ( uint y = startY; y < h; ++y )
		{
			QString playerNumber = _board->playerNumber( x, y );
			QPixmap pixmap = _cache->pixmap( "pics/node_" + playerNumber + ".png" );
			p->drawPixmap( x * NODE_WIDTH, y * NODE_HEIGHT, pixmap );
			
			drawHConnect( p, x, y, playerNumber );
			drawVConnect( p, x, y, playerNumber );
			
			p->save();
			p->translate( x * NODE_WIDTH + NODE_WIDTH/2, y * NODE_HEIGHT + NODE_HEIGHT/2 );
			switch ( _board->at( x, y )->direction().name() )
			{
			case Direction::Down: p->rotate( 90.0 ); break;
			case Direction::Left: p->rotate( 180.0 ); break;
			case Direction::Up: p->rotate( 270.0 ); break;
			default: break;
			}
			p->drawPixmap( -3, -3, arrow );
			p->restore();
		}
	}
	
	// draw focus
	if ( _showSel )
		p->drawWinFocusRect( _selX * NODE_WIDTH, _selY * NODE_HEIGHT, NODE_WIDTH, NODE_HEIGHT, background );
	
	painter.end();
	q->drawPixmap( clipX, clipY, back );
}

void BoardWidget::contentsMouseReleaseEvent( QMouseEvent *e )
{
	hideSelection();
	
	int x = e->x() / NODE_WIDTH;
	int y = e->y() / NODE_HEIGHT;
	if ( x >= 0 && x < _board->width() && y >= 0 && y < _board->height() )
	{
		_selX = x;
		_selY = y;
		emit nodeSelected( x, y );
	}
}

void BoardWidget::boardChanged( int x, int y )
{
	updateContents( x * NODE_WIDTH - NODE_WIDTH/2, y * NODE_HEIGHT - NODE_HEIGHT/2, NODE_WIDTH*2, NODE_HEIGHT*2 );
}

void BoardWidget::ensureNodeVisible( int x, int y )
{
	ensureVisible( x * NODE_WIDTH + NODE_WIDTH/2, y * NODE_HEIGHT + NODE_HEIGHT/2, NODE_WIDTH/2, NODE_HEIGHT/2 );
}

} // namespace Balm

#include "boardwidget.moc"
//kate: indent-mode: csands;
